import urllib2
import json
import time
import ConfigParser

import re
class ip_validator(object):
    @classmethod
    def is_valid_ip(cls, ip):
        """Validates IP addresses.
        """
        return cls.is_valid_ipv4(ip) or cls.is_valid_ipv6(ip)

    @classmethod
    def is_valid_ipv4(cls, ip):
        """Validates IPv4 addresses.
        """
        pattern = re.compile(r"""
        ^
        (?:
          # Dotted variants:
          (?:
            # Decimal 1-255 (no leading 0's)
            [3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}
          |
            0x0*[0-9a-f]{1,2}  # Hexadecimal 0x0 - 0xFF (possible leading 0's)
          |
            0+[1-3]?[0-7]{0,2} # Octal 0 - 0377 (possible leading 0's)
          )
          (?:                  # Repeat 0-3 times, separated by a dot
            \.
            (?:
              [3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}
            |
              0x0*[0-9a-f]{1,2}
            |
              0+[1-3]?[0-7]{0,2}
            )
          ){0,3}
        |
          0x0*[0-9a-f]{1,8}    # Hexadecimal notation, 0x0 - 0xffffffff
        |
          0+[0-3]?[0-7]{0,10}  # Octal notation, 0 - 037777777777
        |
          # Decimal notation, 1-4294967295:
          429496729[0-5]|42949672[0-8]\d|4294967[01]\d\d|429496[0-6]\d{3}|
          42949[0-5]\d{4}|4294[0-8]\d{5}|429[0-3]\d{6}|42[0-8]\d{7}|
          4[01]\d{8}|[1-3]\d{0,9}|[4-9]\d{0,8}
        )
        $
    """, re.VERBOSE | re.IGNORECASE)
        return pattern.match(ip) is not None

    @classmethod
    def is_valid_ipv6(cls, ip):
        """Validates IPv6 addresses.
        """
        pattern = re.compile(r"""
        ^
        \s*                         # Leading whitespace
        (?!.*::.*::)                # Only a single whildcard allowed
        (?:(?!:)|:(?=:))            # Colon iff it would be part of a wildcard
        (?:                         # Repeat 6 times:
            [0-9a-f]{0,4}           #   A group of at most four hexadecimal digits
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
        ){6}                        #
        (?:                         # Either
            [0-9a-f]{0,4}           #   Another group
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
            [0-9a-f]{0,4}           #   Last group
            (?: (?<=::)             #   Colon iff preceeded by exacly one colon
             |  (?<!:)              #
             |  (?<=:) (?<!::) :    #
             )                      # OR
         |                          #   A v4 address with NO leading zeros
            (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            (?: \.
                (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            ){3}
        )
        \s*                         # Trailing whitespace
        $
    """, re.VERBOSE | re.IGNORECASE | re.DOTALL)
        return pattern.match(ip) is not None

class orchestrator(object):
    _name = ""                              # Name of the machine as string
    cluster = "esx-tsd-cluster"
    folder = "autoprovisioned"
    template_name = "w2012r2tpl.uio.no"
    viktighet = 4
    _vlan_id = 0                            # Vlan number as int, this will also change the vlan_name variable see: vlan_id(self, vlan_id)
    vmDiskSize = 0
    vmMemorySize = 2048
    vmNbOfCpus = 1
    vlan_name = "vlan2601-p01"
    _osname = ""                            # Windows here untill we can support linux machines
    datacenter = "TSD"
    ad_domain_id = 1
    ad_domain_name = "tsd.usit.no"
    _ipv4subnetMask = ""                        # Subnet mask as string
    _ipv4Address = ""
    _ipv4gateway = []                           # List of ipv4gateway IPv4 addresses as string
    windows_roles = []
    _ipv4dnsServerList = []                     # List of dns IPv4 addresses as string
    _ipv6dnsServerList = []
    _ipv6gateway = ""
    _ipv6prefixlen = 64
    _ipv6Address = ""
    data = {"parameters": []}
    templates = {}

    config_file = '/etc/tsd/tsd-virtcreate.ini'

    # Where to find the Orchestrator server and services
    authurl        = "https://tsd-vco01.tsd.usit.no:8281"
    orchestrateurl = 'https://tsd-vco01.tsd.usit.no:8281/vco/api/workflows/e7764fb3-d858-4336-af95-2295aca15ce8/executions/'
    getvmsurl      = "https://tsd-vco01.tsd.usit.no:8281/vco/api/workflows/4196069d-b065-4262-bb81-3e7924a5e9b7/executions/"

    def json_add_field(self, value, name, ftype, scope):
        field = {}
        field["name"] = name
        field["scope"] = scope
        field["type"] = ftype

        if ftype == "string":
            if not isinstance(value, str):
                raise Exception("Expected str got ", type(value))
            else:
                field["value"] = {"string": {"value": value}}

        elif ftype == "Array/string":
            if not isinstance(value, list):
                raise Exception("Expected list got", type(value))
            else:
                field["value"] = {"array": {"elements": [{"string" : {"value": i}} for i in value] }}

        elif ftype == "number":
            if isinstance(value, str):
                raise Exception("Expected integer got ", type(value))
            else:
                field["value"] = {"number": {"value": value}}

        self.data["parameters"].append(field)

    def __init__(self, username, password):
        if username is None:
            raise Exception("missing username")
        passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
        passman.add_password(None, self.authurl, username, password)
        authhandler = urllib2.HTTPBasicAuthHandler(passman)
        opener = urllib2.build_opener(authhandler)
        urllib2.install_opener(opener)
        config = ConfigParser.ConfigParser()
        config.read(self.config_file)
        for option in config.options('vmware-templates'):
            self.templates[option] = config.get('vmware-templates', option)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if isinstance(name, basestring):
            self._name = name
            self.json_add_field(self.name, "name", "string", "local")
        else:
            raise Exception("expected str got %s" % type(name))

    @property
    def ipv4gateway(self):
        return self._ipv4gateway

    @ipv4gateway.setter
    def ipv4gateway(self, ipv4gateway):
        if isinstance(ipv4gateway, list):
            for i in ipv4gateway:
                if not ip_validator.is_valid_ip(i):
                    raise Exception("Invalid IP adress ", i)
            self._ipv4gateway = ipv4gateway
            self.json_add_field(self.ipv4gateway, "ipv4gateway", "Array/string", "local")
        else:
            raise Exception("Expected list got ", type(ipv4gateway))

    @property
    def ipv6gateway(self):
        return self._ipv6gateway

    @ipv6gateway.setter
    def ipv6gateway(self, ipv6gateway):
        if ip_validator.is_valid_ipv6(ipv6gateway):
            self._ipv6gateway = ipv6gateway
            self.json_add_field(self.ipv6gateway, "ipv6gateway", "string", "local")
        else:
            raise Exception("Invalid IP adress ", ipv6gateway)

    @property
    def ipv4dnsServerList(self):
        return self._ipv4dnsServerList

    @ipv4dnsServerList.setter
    def ipv4dnsServerList(self, ipv4dnsServerList):
        if isinstance(ipv4dnsServerList, list):
            for i in ipv4dnsServerList:
                if not ip_validator.is_valid_ip(i):
                    raise Exception("Invalid IP adress ", i)
            self._ipv4dnsServerList = ipv4dnsServerList
            self.json_add_field(self.ipv4dnsServerList, "ipv4dnsServerList", "Array/string", "local")
        else:
            raise Exception("Expected list got ", type(ipv4dnsServerList))

    @property
    def ipv6dnsServerList(self):
        return self._ipv6dnsServerList

    @ipv6dnsServerList.setter
    def ipv6dnsServerList(self, ipv6dnsServerList):
        if isinstance(ipv6dnsServerList, list):
            for i in ipv6dnsServerList:
                if not ip_validator.is_valid_ip(i):
                    raise Exception("Invalid IP adress ", i)
            self._ipv6dnsServerList = ipv6dnsServerList
            self.json_add_field(self.ipv6dnsServerList, "ipv6dnsServerList", "Array/string", "local")
        else:
            raise Exception("Expected list got ", type(ipv6dnsServerList))

    @property
    def osname(self):
        return self._osname

    @osname.setter
    def osname(self, osname):
        if osname in ["windows", "linux"]:
            self._osname = osname
            self.template_name = self.templates[osname]
            self.json_add_field(self.osname, "osname", "string", "local")
            self.json_add_field(self.template_name, "template_name", "string", "local")
        else:
            raise Exception("OS name not supported, must be linux or windows", osname)

    @property
    def ipv4Address(self):
        return self._ipv4Address

    @ipv4Address.setter
    def ipv4Address(self, ipv4Address):
        if ip_validator.is_valid_ip(ipv4Address):
            self._ipv4Address = ipv4Address
            self.json_add_field(self.ipv4Address, "ipv4Address", "string", "local")
        else:
            raise Exception("Invalid ip adress ", ipv4Address)

    @property
    def ipv6Address(self):
        return self._ipv6Address

    @ipv6Address.setter
    def ipv6Address(self, ipv6Address):
        if ip_validator.is_valid_ip(ipv6Address):
            self._ipv6Address = ipv6Address
            self.json_add_field(self.ipv6Address, "ipv6Address", "string", "local")
        else:
            raise Exception("Invalid ip adress ", ipv6Address)

    @property
    def ipv4subnetMask(self):
        return self._ipv4subnetMask

    @ipv4subnetMask.setter
    def ipv4subnetMask(self, ipv4subnetMask):
        if ip_validator.is_valid_ip(ipv4subnetMask):
            self._ipv4subnetMask = ipv4subnetMask
            self.json_add_field(self.ipv4subnetMask, "ipv4subnetMask", "string", "local")
        else:
            raise Exception("Invalid ip adress ", ipv4subnetMask)

    @property
    def ipv6prefixlen(self):
        return self._ipv6prefixlen

    @ipv6prefixlen.setter
    def ipv6prefixlen(self, ipv6prefixlen):
        if isinstance(ipv6prefixlen, int):
            self._ipv6prefixlen = ipv6prefixlen
            self.json_add_field(self.ipv6prefixlen, "ipv6prefixlen", "number", "local")
        else:
            raise Exception("Invalid ip adress ", ipv6prefixlen)

    @property
    def vlan_id(self):
        return self._vlan_id

    @vlan_id.setter
    def vlan_id(self, vlan_id):
        if not isinstance(vlan_id, int):
            raise Exception("Expected int got ", type(vlan_id))
        else:
            self._vlan_id = vlan_id
            self.json_add_field(self.vlan_id, "vlan_id", "number", "local")
            # Here we do some string formatting to get the vlan name as
            # represented in vcenter
            self.json_add_field("vlan%d" % vlan_id, "vlan_name", "string", "local")

    def create_vm(self):

        # Add default values to the json datastructure
        self.json_add_field(self.cluster, "cluster", "string", "local")
        self.json_add_field(self.folder, "folder", "string", "local")
        self.json_add_field(self.viktighet, "viktighet", "number", "local")
        self.json_add_field(self.vmDiskSize, "vmDiskSize", "number", "local")
        self.json_add_field(self.vmMemorySize, "vmMemorySize", "number", "local")
        self.json_add_field(self.vmNbOfCpus, "vmNbOfCpus", "number", "local")
        self.json_add_field(self.datacenter, "datacenter", "string", "local")
        self.json_add_field(self.ad_domain_id, "ad_domain_id", "number", "local")
        self.json_add_field(self.ad_domain_name, "ad_domain_name", "string", "local")
        self.json_add_field(self.windows_roles, "windows_roles", "Array/string", "local")
        req = urllib2.Request(self.orchestrateurl)
        req.add_header('Content-Type', 'application/json')
        response = urllib2.urlopen(req, json.dumps(self.data))
        output = dict(response.headers)
        self.state_url = output["location"]
        return self.state_url

    def add_array(self, array):
        elem = []
        for i in array:
            elem.append({"string": {"value": i}})
        return elem

    def check_state(self, url):
        status = json.load(urllib2.urlopen(url))
        vm_state = {}
        vm_state["state"] = (status["state"])
        if "business-state" in status.keys():
            vm_state["progress"] = status["business-state"]
        if "content-exception" in status.keys():
            vm_state["exception"] = status["content-exception"]
        return vm_state

    def get_vms(self):
        json_get_machines = {"parameters": [
            {"value": {"string": {"value": ".*tsd.usit.no"}},
                "type": "string",
                "name": "criteria", }, ]}
        req = urllib2.Request(self.getvmsurl)
        req.add_header('Content-Type', 'application/json')
        response = urllib2.urlopen(req, json.dumps(json_get_machines))

        output = dict(response.headers)
        status = json.load(urllib2.urlopen(output["location"]))

        while (status["state"] in "running"):
            status = self.check_state(output["location"])
            time.sleep(1)

        status = json.load(urllib2.urlopen(output["location"]))
        vm_names = []
        for i in (status["output-parameters"][0]["value"]["array"]["elements"]):
            vm_names.append(i["string"]["value"])
        return vm_names

def test_creation(fqdn):
    print "Trying to create VM %s" % fqdn
    username = None
    password = None
    with open("vmware-virtprov-authfile", "r") as auth:
        for line in auth:
            param, value = line.rstrip().split('=')
            if (param == 'username'):
                username = value
            if (param == 'password'):
                password = value

    if password is None:
        import getpass
        password = getpass.getpass("VMWare password for %s: " % username)

    vm = orchestrator(username, password)
    vm.name = fqdn

    for vmname in vm.get_vms():
        if fqdn == vmname:
            print "VM %s already exist, not creating." % fqdn
            return
        print "VM %s already exist." % vmname

    vm.vlan_id = [2601, 2611][0] # FIXME How should multiple vlans be handled?
    vm.osname = "windows"
    vm.ipv4subnetMask = "255.255.0.0"
    vm.ipv4Address = "10.1.2.3"
    vm.ipv4gateway = ['10.1.0.1']
    vm.ipv4dnsServerList = ['10.1.0.2', '10.1.0.3']
    vm.ipv6Address = "2001:700:111:1::26c"
    vm.ipv6prefixlen = 64
    vm.ipv6dnsServerList = ["2001:700:111:1::2", "2001:700:111:1::3"]
    vm.ipv6gateway = "2001:700:111:1::1"

    ref = vm.create_vm()
    print "Created vm %s, ref %s" % (fqdn, ref)

if __name__ == "__main__":
    test_creation("testmachine.tsd.usit.no")
